"""
This module provides view mappers for running views in asyncio.
"""
import asyncio

import greenlet
from pyramid.config.views import DefaultViewMapper
from pyramid.exceptions import ConfigurationError

from .helpers import run_in_greenlet, is_generator


class AsyncioMapperBase(DefaultViewMapper):
    """
    Base class for asyncio view mappers.
    """

    def run_in_coroutine_view(self, view):

        def coroutine_view(context, request):
            this = greenlet.getcurrent()
            future = asyncio.Future()
            sub_task = asyncio.async(
                run_in_greenlet(this, future, view, context, request)
            )
            while not future.done():
                this.parent.switch(sub_task)
            return future.result()

        return coroutine_view

    def run_in_executor_view(self, view):

        def executor_view(context, request):
            try:
                # since we are running in a new thread,
                # remove the old wsgi.file_wrapper for uwsgi
                request.environ.pop('wsgi.file_wrapper')
            finally:
                this = greenlet.getcurrent()
                future = asyncio.Future()
                sub_task = asyncio.async(
                    run_in_greenlet(
                        this,
                        future,
                        asyncio.get_event_loop().run_in_executor,
                        None,
                        view,
                        context,
                        request,
                    )
                )
                while not future.done():
                    this.parent.switch(sub_task)
                return future.result()

        return executor_view


class CoroutineMapper(AsyncioMapperBase):

    def __call__(self, view):
        if not asyncio.iscoroutinefunction(view) and is_generator(view):
            view = asyncio.coroutine(view)
        else:
            raise ConfigurationError(
                'Non-coroutine {} mapped to coroutine.'.format(view)
            )

        view = super().__call__(view)
        return self.run_in_coroutine_view(view)


class ExecutorMapper(AsyncioMapperBase):

    def __call__(self, view):
        if asyncio.iscoroutinefunction(view) or is_generator(view):
            raise ConfigurationError(
                'Coroutine {} mapped to executor.'.format(view)
            )
        view = super().__call__(view)
        return self.run_in_executor_view(view)


class CoroutineOrExecutorMapper(AsyncioMapperBase):

    def __call__(self, view):

        if asyncio.iscoroutinefunction(view):
            wrapper = self.run_in_coroutine_view
        elif is_generator(view):
            view = asyncio.coroutine(view)
            wrapper = self.run_in_coroutine_view
        else:
            wrapper = self.run_in_executor_view

        view = super().__call__(view)
        return wrapper(view)



import asyncio
import unittest

import greenlet


class TestIsGenerator(unittest.TestCase):

    def test_regular_yield(self):
        from aiopyramid.helpers import is_generator

        def _sample():
            yield 1
            yield 2

        self.assertTrue(is_generator(_sample))

    def test_yield_from(self):
        from aiopyramid.helpers import is_generator

        def _placeholder():
            yield 6
            return 7

        def _sample():
            yield from _placeholder

        self.assertTrue(is_generator(_sample))

    def test_coroutine(self):
        from aiopyramid.helpers import is_generator

        @asyncio.coroutine
        def _sample():
            return 5

        self.assertTrue(is_generator(_sample))

    def test_false(self):
        from aiopyramid.helpers import is_generator

        def _sample():
            return "plain old function"

        self.assertFalse(is_generator(_sample))


class TestSpawnGreenlet(unittest.TestCase):

    def test_return_direct_result(self):
        from aiopyramid.helpers import spawn_greenlet

        def _return_4():
            return 4

        out = asyncio.get_event_loop().run_until_complete(spawn_greenlet(_return_4))
        self.assertEqual(out, 4)

    def test_switch_direct_result(self):
        from aiopyramid.helpers import spawn_greenlet

        def _switch_4_return_5():
            this = greenlet.getcurrent()
            this.parent.switch(4)
            return 5

        out = asyncio.get_event_loop().run_until_complete(spawn_greenlet(_switch_4_return_5))
        self.assertEqual(out, 4)

    def test_wait_on_future(self):
        from aiopyramid.helpers import spawn_greenlet

        future = asyncio.Future()

        def _switch_future():
            this = greenlet.getcurrent()
            this.parent.switch(future)
            return 5  # This should never get returned

        future.set_result(4)  # the result is already set, but spawn_greenlet will still need to yield from it
        out = asyncio.get_event_loop().run_until_complete(spawn_greenlet(_switch_future))
        self.assertEqual(out, 4)


class TestRunInGreenlet(unittest.TestCase):

    def test_result(self):
        from aiopyramid.helpers import spawn_greenlet
        from aiopyramid.helpers import run_in_greenlet

        @asyncio.coroutine
        def _sample(pass_back):
            return pass_back

        def _greenlet():
            this = greenlet.getcurrent()
            future = asyncio.Future()
            message = 12
            sub_task = asyncio.async(run_in_greenlet(this, future, _sample, message))
            this.parent.switch(sub_task)
            self.assertEqual(future.result(), message)
            return message + 1

        out = asyncio.get_event_loop().run_until_complete(spawn_greenlet(_greenlet))
        self.assertEqual(13, out)

    def test_result_chain(self):
        from aiopyramid.helpers import spawn_greenlet
        from aiopyramid.helpers import run_in_greenlet

        @asyncio.coroutine
        def _sample(pass_back):
            return pass_back

        @asyncio.coroutine
        def _chain(pass_back):
            out = yield from _sample(pass_back)
            self.assertEqual(out, pass_back)
            return out - 1

        def _greenlet():
            this = greenlet.getcurrent()
            future = asyncio.Future()
            message = 12
            sub_task = asyncio.async(run_in_greenlet(this, future, _chain, message))
            this.parent.switch(sub_task)
            self.assertEqual(future.result(), message - 1)
            return message + 1

        out = asyncio.get_event_loop().run_until_complete(spawn_greenlet(_greenlet))
        self.assertEqual(13, out)

    def test_exception(self):
        from aiopyramid.helpers import spawn_greenlet
        from aiopyramid.helpers import run_in_greenlet

        @asyncio.coroutine
        def _sample():
            raise KeyError

        def _greenlet():
            this = greenlet.getcurrent()
            future = asyncio.Future()
            sub_task = asyncio.async(run_in_greenlet(this, future, _sample))
            this.parent.switch(sub_task)
            self.assertRaises(KeyError, future.result)

        asyncio.get_event_loop().run_until_complete(spawn_greenlet(_greenlet))



import asyncio
import inspect

import greenlet


def is_generator(func):
    """ Tests whether `func` is capable of becoming an `asyncio.coroutine`. """
    return (
        inspect.isgeneratorfunction(func) or
        isinstance(func, asyncio.Future) or
        inspect.isgenerator(func)
    )


@asyncio.coroutine
def spawn_greenlet(func, *args):
    """
    Spawns a new greenlet and waits on any `asyncio.Future` objects returned.

    This is used by the Gunicorn worker to proxy a greenlet within an `asyncio`
    event loop.
    """

    g = greenlet.greenlet(func)
    result = g.switch(*args)
    while True:
        if isinstance(result, asyncio.Future):
            result = yield from result
        else:
            break
    return result


@asyncio.coroutine
def run_in_greenlet(back, future, func, *args):
    """
    Wait for :term:`coroutine` func and switch back to the request greenlet
    setting any result in the future or an Exception where approrpiate.

    func is often a :term:`view callable`
    """
    try:
        result = yield from func(*args)
    except Exception as ex:
        future.set_exception(ex)
    else:
        future.set_result(result)
    finally:
        return back.switch()



import greenlet
import asyncio
import logging

log = logging.getLogger(__name__)


def coroutine_logger_tween_factory(handler, registry):
    """
    Example of an asyncronous tween that delegates a syncronous function to
    a child thread. This tween asyncronously logs all requests and responses.
    """

    @asyncio.coroutine
    def _async_log(back, content):
        # log doesn't really need to be run in a separate thread
        # but it works for demonstration purposes
        # NOTE: this does not do exception handling, which means
        # that no 500 error view will be run. The exception needs to
        # be reraised in the request greenlet. See helpers.run_in_greenlet for
        # and example.

        yield from asyncio.get_event_loop().run_in_executor(
            None,
            log.info,
            content
        )
        back.switch()

    def coroutine_logger_tween(request):
        # get this greenlet
        this = greenlet.getcurrent()

        # queue request logging
        sub_task = asyncio.async(_async_log(this, request))

        # switch to parent so that aio loop runs
        # (waiting for subtask if gunicorn)
        this.parent.switch(sub_task)

        # when request is logged, it will switch back to this

        # get response, this could also be done in a coroutine
        response = handler(request)

        # queue respone logging
        sub_task = asyncio.async(_async_log(this, response))

        # switch to parent so that aio loop runs
        # (waiting for subtask if gunicorn)
        this.parent.switch(sub_task)

        # when response is logged, it will switch back to this

        # return response after logging is done
        return response

    return coroutine_logger_tween



import asyncio
import inspect

import websockets
import gunicorn  # noqa

from pyramid.response import Response

from aiopyramid.config import AsyncioMapperBase


class HandshakeInterator:

    def __init__(self, app_iter):
        self.content = list(app_iter)
        self.index = 0

    def __iter__(self):
        return self

    def __next__(self):
        try:
            return self.content[self.index]
        except IndexError:
            raise StopIteration
        finally:
            self.index += 1


class SwitchProtocolsResponse(Response):
    """Upgrade from a WSGI connection with the WebSocket handshake."""

    def __init__(self, environ, switch_protocols):
        super().__init__()
        self.status_int = 101

        http_1_1 = environ['SERVER_PROTOCOL'] == 'HTTP/1.1'
        get_header = lambda k: environ['HTTP_' + k.upper().replace('-', '_')]
        key = websockets.handshake.check_request(get_header)

        if not http_1_1 or key is None:
            self.status_int = 400
            self.content = "Invalid WebSocket handshake.\n"
        else:
            set_header = self.headers.__setitem__
            websockets.handshake.build_response(set_header, key)
            self.app_iter = HandshakeInterator(self.app_iter)
            self.app_iter.close = switch_protocols


class WebsocketMapper(AsyncioMapperBase):

    def launch_websocket_view(self, view):

        def websocket_view(context, request):

            if inspect.isclass(view):
                view_callable = view(context, request)
            else:
                view_callable = view

            @asyncio.coroutine
            def _ensure_ws_close(ws):
                yield from view_callable(ws)
                yield from ws.close()

            def switch_protocols():
                ws_protocol = websockets.WebSocketCommonProtocol()
                transport = request.environ['async.writer'].transport
                transport._protocol = ws_protocol
                ws_protocol.connection_made(transport)
                asyncio.async(_ensure_ws_close(ws_protocol))

            response = SwitchProtocolsResponse(request.environ, switch_protocols)
            # convert iterator to avoid eof issues
            response.body = response.body

            return response

        return websocket_view

    def __call__(self, view):
        """ Accepts a view_callable class. """
        return self.launch_websocket_view(view)
